@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2005, 2006,
- *	2007, 2008 Tama Communications Corporation
+ *	2007, 2008, 2010 Tama Communications Corporation
  *
  * This file is part of GNU GLOBAL.
  *
@@ -41,19 +41,21 @@
 #include "getopt.h"
 
 #include "global.h"
+#include "parser.h"
 #include "regex.h"
 #include "const.h"
 
 static void usage(void);
 static void help(void);
 static void setcom(int);
-int decide_tag_by_context(const char *, const char *, const char *);
+int decide_tag_by_context(const char *, const char *, int);
 int main(int, char **);
 void completion(const char *, const char *, const char *);
 void idutils(const char *, const char *);
 void grep(const char *, const char *);
 void pathlist(const char *, const char *);
-void parsefile(int, char **, const char *, const char *, const char *, int);
+void parsefile(int, char *const *, const char *, const char *, const char *, int);
+void parsefile_using_builtin_parser(char *const *, const char *, const char *, const char *, int);
 int search(const char *, const char *, const char *, const char *, int);
 void tagsearch(const char *, const char *, const char *, const char *, int);
 void encode(char *, int, const char *);
@@ -166,13 +168,18 @@ setcom(int c)
  *	i)	lineno	context lineno
  *	r)		GTAGS, GRTAGS, GSYMS
  */
+#define NEXT_NUMBER(p) do {                                                         \
+	for (n = 0; isdigit(*p); p++)                                               \
+		n = n * 10 + (*p - '0');                                            \
+} while (0)
 int
-decide_tag_by_context(const char *tag, const char *file, const char *lineno)
+decide_tag_by_context(const char *tag, const char *file, int lineno)
 {
 	char path[MAXPATHLEN+1], s_fid[32];
 	const char *tagline, *p;
 	DBOP *dbop;
 	int db = GSYMS;
+	int iscompline = 0;
 
 	if (normalize(file, get_root_with_slash(), cwd, path, sizeof(path)) == NULL)
 		die("'%s' is out of source tree.", file);
@@ -189,6 +196,8 @@ decide_tag_by_context(const char *tag, const char *file, const char *lineno)
 	 * read btree records directly to avoid the overhead.
 	 */
 	dbop = dbop_open(makepath(dbpath, dbname(GTAGS), NULL), 0, 0, 0);
+	if (dbop_getoption(dbop, COMPLINEKEY))
+		iscompline = 1;
 	tagline = dbop_first(dbop, tag, NULL, 0);
 	if (tagline)
 		db = GTAGS;
@@ -200,14 +209,52 @@ decide_tag_by_context(const char *tag, const char *file, const char *lineno)
 		if (p != NULL && *p == ' ') {
 			for (p++; *p && *p != ' '; p++)
 				;
-			if (*p++ != ' ')
-				die("Impossible!");
-			if ((p = locatestring(p, lineno, MATCH_AT_FIRST)) != NULL && *p == ' ') {
-				db = GRTAGS;
-				break;
+			if (*p++ != ' ' || !isdigit(*p))
+				die("Impossible! decide_tag_by_context(1)");
+			/*
+			 * Standard format	n <blank> <image>$
+			 * Compact format	d,d,d,d$
+			 */
+			if (!iscompline) {			/* Standard format */
+				if (atoi(p) == lineno) {
+					db = GRTAGS;
+					goto finish;
+				}
+			} else {				/* Compact format */
+				int n, cur, last = 0;
+
+				do {
+					if (!isdigit(*p))
+						die("Impossible! decide_tag_by_context(2)");
+					NEXT_NUMBER(p);
+					cur = last + n;
+					if (cur == lineno) {
+						db = GRTAGS;
+						goto finish;
+					}
+					last = cur;
+					if (*p == '-') {
+						if (!isdigit(*++p))
+							die("Impossible! decide_tag_by_context(3)");
+						NEXT_NUMBER(p);
+						cur = last + n;
+						if (lineno >= last && lineno <= cur) {
+							db = GRTAGS;
+							goto finish;
+						}
+						last = cur;
+					}
+					if (*p) {
+						if (*p == ',')
+							p++;
+						else
+							die("Impossible! decide_tag_by_context(4)");
+					}
+				} while (*p);
 			}
 		}
 	}
+finish:
         dbop_close(dbop);
 	return db;
 }
@@ -479,7 +526,7 @@ main(int argc, char **argv)
 	if (context_file) {
 		if (isregex(av))
 			die_with_code(2, "regular expression is not allowed with the --from-here option.");
-		db = decide_tag_by_context(av, context_file, context_lineno);
+		db = decide_tag_by_context(av, context_file, atoi(context_lineno));
 	} else {
 		db = (rflag) ? GRTAGS : ((sflag) ? GSYMS : GTAGS);
 	}
@@ -530,7 +577,10 @@ main(int argc, char **argv)
 	 * parse source files.
 	 */
 	else if (fflag) {
-		parsefile(argc, argv, cwd, root, dbpath, db);
+		if (getconfb("use_builtin_parser"))
+			parsefile_using_builtin_parser(argv, cwd, root, dbpath, db);
+		else
+			parsefile(argc, argv, cwd, root, dbpath, db);
 	}
 	/*
 	 * tag search.
@@ -865,7 +915,7 @@ pathlist(const char *pattern, const char *dbpath)
  *	i)	db	type of parse
  */
 void
-parsefile(int argc, char **argv, const char *cwd, const char *root, const char *dbpath, int db)
+parsefile(int argc, char *const *argv, const char *cwd, const char *root, const char *dbpath, int db)
 {
 	CONVERT *cv;
 	int count = 0;
@@ -905,7 +955,7 @@ parsefile(int argc, char **argv, const char *cwd, const char *root, const char *
 		char path[MAXPATHLEN+1];
 
 		/*
-		 * convert the path into relative from the root directory of source tree.
+		 * convert the path into relative to the root directory of source tree.
 		 */
 		if (normalize(av, get_root_with_slash(), cwd, path, sizeof(path)) == NULL)
 			if (!qflag)
@@ -980,6 +1030,151 @@ parsefile(int argc, char **argv, const char *cwd, const char *root, const char *
 		fprintf(stderr, " (no index used).\n");
 	}
 }
+/*
+ * parsefile_using_builtin_parser: parse file to pick up tags.
+ *
+ *	i)	argv
+ *	i)	cwd	current directory
+ *	i)	root	root directory of source tree
+ *	i)	dbpath	dbpath
+ *	i)	db	type of parse
+ */
+#define TARGET_DEF	(1 << GTAGS)
+#define TARGET_REF	(1 << GRTAGS)
+#define TARGET_SYM	(1 << GSYMS)
+struct parsefile_data {
+	CONVERT *cv;
+	DBOP *dbop;
+	int target;
+	int extractmethod;
+	int count;
+};
+static void
+put_syms(int type, const char *tag, int lno, const char *path, const char *line_image, void *arg)
+{
+	struct parsefile_data *data = arg;
+	const char *key;
+
+	if (format == FORMAT_PATH && data->count > 0)
+		return;
+	switch (type) {
+	case PARSER_DEF:
+		if (!(data->target & TARGET_DEF))
+			return;
+		break;
+	case PARSER_REF_SYM:
+		if (!(data->target & (TARGET_REF | TARGET_SYM)))
+			return;
+		/*
+		 * extract method when class method definition.
+		 *
+		 * Ex: Class::method(...)
+		 *
+		 * key	= 'method'
+		 * data = 'Class::method  103 ./class.cpp ...'
+		 */
+		if (data->extractmethod) {
+			if ((key = locatestring(tag, ".", MATCH_LAST)) != NULL)
+				key++;
+			else if ((key = locatestring(tag, "::", MATCH_LAST)) != NULL)
+				key += 2;
+			else
+				key = tag;
+		} else {
+			key = tag;
+		}
+		if (dbop_get(data->dbop, key) != NULL) {
+			if (!(data->target & TARGET_REF))
+				return;
+		} else {
+			if (!(data->target & TARGET_SYM))
+				return;
+		}
+		break;
+	default:
+		return;
+	}
+	convert_put_using(data->cv, tag, path, lno, line_image);
+	data->count++;
+}
+void
+parsefile_using_builtin_parser(char *const *argv, const char *cwd, const char *root, const char *dbpath, int db)
+{
+	int count = 0;
+	int file_count = 0;
+	STRBUF *sb = strbuf_open(0);
+	const char *langmap, *av;
+	char path[MAXPATHLEN+1];
+	struct parsefile_data data;
+
+	data.target = 1 << db;
+	data.extractmethod = getconfb("extractmethod");
+	if (getconfs("langmap", sb))
+		langmap = strbuf_value(sb);
+	else
+		langmap = DEFAULTLANGMAP;
+	data.cv = convert_open(type, format, root, cwd, dbpath, stdout);
+	if (gpath_open(dbpath, 0) < 0)
+		die("GPATH not found.");
+	if (data.target & (TARGET_REF | TARGET_SYM)) {
+		data.dbop = dbop_open(makepath(dbpath, dbname(GTAGS), NULL), 0, 0, 0);
+		if (data.dbop == NULL)
+			die("%s not found.", dbname(GTAGS));
+	} else {
+		data.dbop = NULL;
+	}
+	/*
+	 * Execute parser in the root directory of source tree.
+	 */
+	if (chdir(root) < 0)
+		die("cannot move to '%s' directory.", root);
+	parser_init(langmap, NULL);
+	while ((av = *argv++) != NULL) {
+		/*
+		 * convert the path into relative to the root directory of source tree.
+		 */
+		if (normalize(av, get_root_with_slash(), cwd, path, sizeof(path)) == NULL)
+			if (!qflag)
+				fprintf(stderr, "'%s' is out of source tree.\n", path + 2);
+		if (!gpath_path2fid(path, NULL)) {
+			if (!qflag)
+				fprintf(stderr, "'%s' not found in GPATH.\n", path + 2);
+			continue;
+		}
+		if (!test("f", makepath(root, path, NULL))) {
+			if (test("d", NULL)) {
+				if (!qflag)
+					fprintf(stderr, "'%s' is a directory.\n", av);
+			} else {
+				if (!qflag)
+					fprintf(stderr, "'%s' not found.\n", av);
+			}
+			continue;
+		}
+		if (lflag && !locatestring(path, localprefix, MATCH_AT_FIRST))
+			continue;
+		data.count = 0;
+		parse_file(path, 0, put_syms, &data);
+		count += data.count;
+		file_count++;
+	}
+	if (chdir(cwd) < 0)
+		die("cannot move to '%s' directory.", cwd);
+	/*
+	 * Settlement
+	 */
+	if (data.dbop != NULL)
+		dbop_close(data.dbop);
+	gpath_close();
+	convert_close(data.cv);
+	strbuf_close(sb);
+	if (file_count == 0)
+		die("file not found.");
+	if (vflag) {
+		print_count(count);
+		fprintf(stderr, " (no index used).\n");
+	}
+}
 /*
  * search: search specified function 
  *
