@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2006
+ * Copyright (c) 2006, 2008
  *	Tama Communications Corporation
  *
  * This file is part of GNU GLOBAL.
@@ -55,7 +55,7 @@ static char *include_pattern(const char *);
 static void command_help(void);
 static void command_loop(void);
 static int execute_command(STRBUF *, const int, const int, const char *);
-static void search(int, const char *);
+static void search(int, char *);
 
 int show_version;
 int show_help;
@@ -63,6 +63,7 @@ int qflag;
 int vflag;
 
 #define NA	-1
+#define FROM_HERE -2
 
 /*
 static void
@@ -90,7 +91,7 @@ static struct option const long_options[] = {
 };
 
 char global_path[MAXFILLEN];
-
+char *context;
 int ignore_case;
 
 /*
@@ -127,7 +128,7 @@ main(int argc, char **argv)
 	int optchar;
 	int option_index = 0;
 
-	while ((optchar = getopt_long(argc, argv, "Cqv", long_options, &option_index)) != EOF) {
+	while ((optchar = getopt_long(argc, argv, "bCcdeF:f:hI:i:kLlp:qRs:TUuVv0123456789", long_options, &option_index)) != EOF) {
 		switch (optchar) {
 		case 0:
 			break;
@@ -141,7 +142,6 @@ main(int argc, char **argv)
 			vflag++;
 			break;
 		default:
-			/* usage(); */
 			break;
 		}
 	}
@@ -330,12 +330,15 @@ execute_command(STRBUF *sb, const int com, const int opt, const char *arg)
 	strbuf_puts(command, global_path);
 	strbuf_puts(command, " --result=cscope");
 	if (opt || ignore_case) {
-		strbuf_putc(command, ' ');
-		strbuf_putc(command, '-');
-		if (opt)
+		if (opt == FROM_HERE) {
+			strbuf_puts(command, " --from-here=");
+			strbuf_puts(command, context);
+		} else {
+			strbuf_puts(command, " -");
 			strbuf_putc(command, opt);
+		}
 		if (ignore_case)
-			strbuf_putc(command, 'i');
+			strbuf_puts(command, " --ignore-case");
 	}
 	strbuf_putc(command, ' ');
 	strbuf_putc(command, QUOTE);
@@ -408,10 +411,10 @@ execute_command(STRBUF *sb, const int com, const int opt, const char *arg)
  * Unsupported command prints "cscope: 0 lines\n".
  */
 static void
-search(int com, const char *arg)
+search(int com, char *arg)
 {
 	static STRBUF *sb;
-	char buf[1024];
+	char buf[1024], *p;
 	int count = 0, opt = 0;
 
 	if (sb == NULL)
@@ -426,7 +429,14 @@ search(int com, const char *arg)
 	case '1':		/* Find this definition */
 		break;
 	case '2':		/* Find functions called by this function */
-		opt = NA;
+		/*
+		 * <symbol>:<line number>:<path>
+		 */
+		for (p = arg; *p && *p != ':'; p++)
+			;
+		*p++ = '\0';
+		context = p;
+		opt = FROM_HERE;
 		break;
 	case '3':		/* Find functions calling this function */
 		opt = 'r';
