@@ -1,7 +1,7 @@
 ;;; gtags.el --- gtags facility for Emacs
 
 ;;
-;; Copyright (c) 1997, 1998, 1999, 2000, 2006, 2007, 2008
+;; Copyright (c) 1997, 1998, 1999, 2000, 2006, 2007, 2008, 2009, 2010
 ;;	Tama Communications Corporation
 ;;
 ;; This file is part of GNU GLOBAL.
@@ -22,7 +22,7 @@
 
 ;; GLOBAL home page is at: http://www.gnu.org/software/global/
 ;; Author: Tama Communications Corporation
-;; Version: 2.5
+;; Version: 2.6
 ;; Keywords: tools
 ;; Required version: GLOBAL 5.7 or later
 
@@ -92,6 +92,11 @@
   :group 'gtags
   :type 'boolean)
 
+(defcustom gtags-select-buffer-single nil
+  "*If non-nil, gtags select buffer is single."
+  :group 'gtags
+  :type 'boolean)
+
 ;; Variables
 (defvar gtags-current-buffer nil
   "Current buffer.")
@@ -223,21 +228,38 @@
   (gtags-completing 'gtags string predicate code))
 (defun gtags-completing-gsyms (string predicate code)
   (gtags-completing 'gsyms string predicate code))
+(defun gtags-completing-files (string predicate code)
+  (gtags-completing 'files string predicate code))
 ;; common part of completing-XXXX
-;;   flag: 'gtags or 'gsyms
+;;   flag: 'gtags or 'gsyms or 'files
 (defun gtags-completing (flag string predicate code)
-  (let ((option "-c")
+  ; The purpose of using the -n option for the -P command is to exclude
+  ; dependence on the execution directory.
+  (let ((option (cond ((eq flag 'files) "-Pon")
+                      ((eq flag 'gsyms)  "-cs")
+                      (t                "-c")))
         (complete-list (make-vector 63 0))
         (prev-buffer (current-buffer)))
     ; build completion list
     (set-buffer (generate-new-buffer "*Completions*"))
-    (if (eq flag 'gsyms)
-        (setq option (concat option "s")))
     (call-process "global" nil t nil option string)
     (goto-char (point-min))
-    (while (looking-at gtags-symbol-regexp)
-      (intern (gtags-match-string 0) complete-list)
-      (forward-line))
+    ;
+    ; The specification of the completion for files is different from that for symbols.
+    ; The completion for symbols matches only to the head of the symbol. But the completion
+    ; for files matches any part of the path.
+    ;
+    (if (eq flag 'files)
+        ; extract input string and the following part.
+        (let ((match-string (if (equal "" string) "\./\\(.*\\)" (concat ".*\\(" string ".*\\)"))))
+          (while (not (eobp))
+            (looking-at match-string)
+            (intern (gtags-match-string 1) complete-list)
+            (forward-line)))
+      (while (not (eobp))
+        (looking-at gtags-symbol-regexp)
+        (intern (gtags-match-string 0) complete-list)
+        (forward-line)))
     (kill-buffer (current-buffer))
     ; recover current buffer
     (set-buffer prev-buffer)
@@ -349,10 +371,11 @@
   (interactive)
   (let (tagname prompt input)
     (setq prompt "Find files: ")
-    (setq input (read-string prompt))
+    (setq input (completing-read prompt 'gtags-completing-files
+                  nil nil nil gtags-history-list))
     (if (not (equal "" input)) (setq tagname input))
     (gtags-push-context)
-    (gtags-goto-tag tagname "P")))
+    (gtags-goto-tag tagname "Po")))
 
 (defun gtags-parse-file ()
   "Input file name, parse it and show object list."
@@ -469,28 +492,52 @@
 
 ;; goto tag's point
 (defun gtags-goto-tag (tagname flag &optional other-win)
-  (let (option context save prefix buffer lines)
+  (let (option context save prefix buffer lines flag-char)
     (setq save (current-buffer))
+    (setq flag-char (string-to-char flag))
     ; Use always ctags-x format.
     (setq option "-x")
-    (if (equal flag "C")
+    (if (char-equal flag-char ?C)
         (setq context (concat "--from-here=" (number-to-string (gtags-current-lineno)) ":" buffer-file-name))
         (setq option (concat option flag)))
     (cond
-     ((equal flag "C")
+     ((char-equal flag-char ?C)
       (setq prefix "(CONTEXT)"))
-     ((equal flag "P")
+     ((char-equal flag-char ?P)
       (setq prefix "(P)"))
-     ((equal flag "g")
+     ((char-equal flag-char ?g)
       (setq prefix "(GREP)"))
-     ((equal flag "I")
+     ((char-equal flag-char ?I)
       (setq prefix "(IDUTILS)"))
-     ((equal flag "s")
+     ((char-equal flag-char ?s)
       (setq prefix "(S)"))
-     ((equal flag "r")
+     ((char-equal flag-char ?r)
       (setq prefix "(R)"))
      (t (setq prefix "(D)")))
     ;; load tag
+    (if gtags-select-buffer-single
+        (progn
+          ; delete "*GTAGS SELECT*" buffer info from gtags-buffer-stack and gtags-point-stack
+          (let (now-gtags-buffer-stack now-buffer now-gtags-point-stack now-point)
+            (setq now-gtags-buffer-stack gtags-buffer-stack)
+            (setq now-gtags-point-stack gtags-point-stack)
+            (while now-gtags-buffer-stack
+              (setq now-buffer (car now-gtags-buffer-stack))
+              (setq now-point (car now-gtags-point-stack))
+              (if (string-match "*GTAGS SELECT*" (buffer-name now-buffer))
+                  (progn
+                    (delete now-buffer gtags-buffer-stack)
+                    (delete now-point gtags-point-stack)))
+              (setq now-gtags-buffer-stack (cdr now-gtags-buffer-stack))
+              (setq now-gtags-point-stack (cdr now-gtags-point-stack))))
+          ; kill "*GTAGS SELECT*" buffer
+          (let (now-buffer-list now-buffer)
+            (setq now-buffer-list (buffer-list))
+            (while now-buffer-list
+              (setq now-buffer (car now-buffer-list))
+              (if (string-match "*GTAGS SELECT*" (buffer-name now-buffer))
+                  (kill-buffer now-buffer))
+              (setq now-buffer-list (cdr now-buffer-list))))))
     (setq buffer (generate-new-buffer (generate-new-buffer-name (concat "*GTAGS SELECT* " prefix tagname))))
     (set-buffer buffer)
     ;
@@ -519,13 +566,13 @@
       (cond
        ((= 0 lines)
          (cond
-          ((equal flag "P")
+          ((char-equal flag-char ?P)
            (message "%s: path not found" tagname))
-          ((equal flag "g")
+          ((char-equal flag-char ?g)
            (message "%s: pattern not found" tagname))
-          ((equal flag "I")
+          ((char-equal flag-char ?I)
            (message "%s: token not found" tagname))
-          ((equal flag "s")
+          ((char-equal flag-char ?s)
            (message "%s: symbol not found" tagname))
           (t
            (message "%s: tag not found" tagname)))
