@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2005, 2006, 2007
+ * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2005, 2006, 2007, 2010
  *	Tama Communications Corporation
  *
  * This file is part of GNU GLOBAL.
@@ -20,8 +20,8 @@
 
 #ifndef _GTOP_H_
 #define _GTOP_H_
+
 #include <stdio.h>
-#include <ctype.h>
 
 #include "gparam.h"
 #include "dbop.h"
@@ -101,15 +101,21 @@ typedef struct {
 	 */
 	char cur_path[MAXPATHLEN+1];	/* current path */
 	STRBUF *sb;			/* string buffer */
-	FILE *fp;			/* descriptor of 'path' */
 	/* used for compact format and path name only read */
 	STRHASH *path_hash;
-
+	/*
+	 * Stuff for 1-pass parsing
+	 */
+	FILE *fp;
 } GTOP;
 
 const char *dbname(int);
-GTOP *gtags_open(const char *, const char *, int, int);
+GTOP *gtags_open(const char *, const char *, int, int, int);
 void gtags_put(GTOP *, const char *, const char *);
+void gtags_put_using(GTOP *, const char *, int, const char *, const char *);
+void gtags_flush(GTOP *, const char *);
+void gtags_add_ref_sym(GTOP *const *, FILE *);
+void gtags_move_ref_sym(GTOP *const *);
 void gtags_delete(GTOP *, IDSET *);
 GTP *gtags_first(GTOP *, const char *, int);
 GTP *gtags_next(GTOP *);
